function u = l1nonlinfit_ssn(fem,Mf,yd,alpha,ue,opts)
% U = L1NONLINFIT_SSN(FEM,MF,YD,ALPHA,UE,OPTS)
% solve nonlinear L1 data fitting problem using a path-following semi-smooth
%  Newton method
%
% Input:  fem    -- finite element structure containing matrices
%                   (A(u),M,Mx,Y), projection operators (P), regularization
%                   term (DD)
%         Mf     -- right hand side with mass matrix already applied
%         yd     -- measured data
%         alpha  -- regularization parameter for L2 penalty
%         ue     -- exact solution (to evaluate reconstruction error)
%         opts   -- options structure containing (defaults)
%            beta0 -- initial penalty parameter for p (1)
%            bred  -- reduction factor for beta (1/2)
%            bmin  -- minimal beta value (1e-12)
%            c     -- penalty parameter for box constraint (1e3)
%            maxit -- max. number of semi-smooth Newton iterations (20)
%            tol   -- tolerance for Newton iteration: norm of gradient (1e-6)
%            cgits -- max. number of BiCGstab iterations (1000)
%            cgtol -- tolerance for BiCGstab (1e-6)
%            verb  -- verbosity (0,1: no display, 2: only path-following, 3: also SSN)
% Output: u      -- reconstruction
%
% Christian Clason (christian.clason@uni-graz.at)
% Bangti Jin       (btjin@math.tamu.edu)
% February 27, 2011

%% Setup parameters
nel   = length(fem.x);      % problem size: number of nodes

% Default parameters
if nargin <= 5 || isempty(opts)
    beta  = 1;              % initial penalty parameter for p
    bred  = 1/2;            % reduction factor for beta
    bmin  = 1e-12;          % minimal beta value
    maxit = 20;             % max. number of semi-smooth Newton iterations
    tol   = 1e-6;           % tolerance for Newton iteration (norm of gradient)
    cgits = 1000;           % max. number of BiCGstab iterations
    cgtol = 1e-6;           % tolerance for BiCGstab
    verb  = 1;              % verbosity (0,1: no display, 2: only path-following, 3: also SSN)
else
    beta  = opts.beta;
    bred  = opts.bred;
    bmin  = opts.bmin;
    maxit = opts.maxit;
    tol   = opts.tol;
    cgits = opts.cgits;
    cgtol = opts.cgtol;
    verb  = opts.verb;
end
DD = fem.DD;     % regularization term
M  = fem.M0;     % Mass matrix
Mx = fem.Mx;     % differentiated mass matrix
Y  = fem.Y;      % Y(v)*du assembles weighted mass matrix for piecewise constant du and linear v
P  = fem.P;      % projection from U_h to V_h

%% SSN loop

% initialize iterates
u = ones(nel,1);

while beta>bmin
    
    iter   = 0;
    update = 1;
    nres   = 1;
    uold   = u;
    
    while (iter<maxit)
        
        A   = fem.A(u);           % differential operator
        y   = A\(Mf);             % state
        res = y-yd;               % residual
        
        % active sets
        Ap = (res - beta)>0;      
        Am = (res + beta)<0;      
        As = Ap+Am;
        
        p = (1-As).*res/beta + (Ap - Am);
        
        q = A\(M*p);              % adjoint
        
        % precompute mass matrices for Hessian application
        yx = Mx*y; Myx = Y(yx);
        qx = Mx*q; Mqx = Y(qx);
        
        % gradient (preconditioning with DD^(-1))
        rhs  = -(alpha*u - DD\(P(yx.*qx)));
        nres = norm(rhs)/sqrt(nel);
        
        % Hessian
        C = @(x) applyHess(x,A,DD,M,Mx,P,yx,qx,Myx,Mqx,As,alpha,beta);
        
        % Newton step
        [dx,flag] = bicgstab(C,rhs,cgtol,cgits); %#ok<NASGU>
        
        u = u + dx;
        
        if iter == 0
            update = 1;
        else
            update = nnz((Ap-Ap_old)|(Am-Am_old));
            if verb > 2
                fprintf('SSN iteration %d: %d points changed in active set, beta=%1.2e, |grad|=%e\n', iter, update,beta,nres);
                
            end
        end
        iter   = iter+1;
        Ap_old = Ap;
        Am_old = Am;
        
        if ~update && nres<tol    % terminate if active sets coincide and norm of gradient small
            break
        end
        
    end
    
    if verb > 1
        fprintf('beta=%1.2e: %d points changed in active set after %d iterations, |grad| = %e, |u-uex|=%e \n',beta,update,iter,nres,norm(u-ue)/sqrt(nel));
    end
    
    if update || nres > 1         % iterate not feasible: terminate and return previous iterate
        u = uold;
        break
    else                          % reduce beta, continue
        beta = beta*bred;
    end
    
end

end % main function


%% Application of Hessian
function Hdu = applyHess(x,A,DD,M,Mx,P,yx,qx,Myx,Mqx,As,alpha,beta)
dy = A\(Myx*x);                   % differential change in state
dp = (1-As).*dy/beta;
dq = A\(Mqx*x-(M*dp));            % differential change in adjoint

Hdu = alpha*x - DD\(P(yx.*(Mx*dq)+qx.*(Mx*dy)));
end % function applyHess
