% L1NONLINFIT_TEST2D
% Test script for nonlinear L^1 data fitting using automatic parameter
% choice and path-following semismooth Newton method (2D). Solves
%  min_u |S(u)-yd|_L1 + alpha/2 (|u|_H1)^2, where S(u)=:y satisfies
%   <u \nabla y, \nabla v > = <f,v> for all v in H^1
%
% Christian Clason (christian.clason@uni-graz.at)
% Bangti Jin       (btjin@math.tamu.edu)
% February 27, 2011

clear, close all
%% parameters
n     = 64;        % problem size
d_per = 0.3;       % percentage of corrupted data points
verb  = 2;         % verbosity (0: no display, 1: balancing only, 2: also path-following, 3: also SSN)

% ... for balancing principle
alpha0 = 1e-1;     % initial guess for parameter alpha
sigma  = 1.001;    % balancing factor
tol    = 1e-4;     % terminate if relative change in alpha less than tol

% ... for path-following semi-smooth Newton method
opts.beta  = 1;              % initial penalty parameter for p
opts.bred  = 0.5;            % reduction factor for beta
opts.bmin  = 1e-12;          % minimal beta value
opts.maxit = 20;             % max. number of semi-smooth Newton iterations
opts.tol   = 1e-6;           % tolerance for Newton iteration (norm of gradient)
opts.cgits = 100;            % max. number of BiCGstab iterations
opts.cgtol = 1e-6;           % tolerance for BiCGstab
opts.verb  = verb;           % verbosity

%% Data
fem = setupFEM2d(n);         % FEM setup
xx  = fem.xx; yy = fem.yy;   % grid
Mf  = fem.M0*(1+0*xx(:));    % right hand side (pre-multiply mass matrix)
ue =  1/2+xx.^2.*yy.^2;      % true solution
ye  = fem.A(ue)\(Mf);        % exact data: cell-wise by projection

% noisy data (additive random-valued impulsive noise)
nel    = fem.nel; nnod = length(xx(:)); % number of element, nodes
y      = ye;
ind    = find(rand(nnod,1) < d_per);
y(ind) = ye(ind) + max(abs(ye)) * randn(length(ind),1);
delta  = norm(y-ye,1)/nnod;  % compute noise level

% show the true solution and noisy data
figure(1); fem.plot(ue);title('exact solution');
shading interp; camlight headlight; lighting phong; drawnow;
figure(2); fem.plot(ye);title('exact data');
shading interp; camlight headlight; lighting phong; drawnow;
figure(3); fem.plot(y);title('noisy data');
shading interp; camlight headlight; lighting phong; drawnow;

%% Balancing principle loop
t_b = tic;

alpha = alpha0;
for iter = 1:10
    alpha1 = alpha;
    
    % compute u_\alpha
    t_s    = tic;
    u      = l1nonlinfit_ssn2d(fem,Mf,y,alpha1,ue,opts);
    t_ssn  = toc(t_s);
    
    % update alpha
    alpha = (sigma-1)*norm(fem.M*(fem.A(u)\Mf-y),1)/((u(:)'*fem.DD*u(:))/2);
    err   = norm(u-ue(:))/sqrt(nnod);
    
    if opts.verb > 0
        figure(99); fem.plot(u);
        shading interp; camlight headlight; lighting phong; axis tight; drawnow;
        display(sprintf('Iteration %d: t = %g seconds, alpha = %g, err= %g',...
            iter,t_ssn,alpha,err));
    end
    
    % terminate if relative change in alpha too small
    if (abs(alpha-alpha1)/alpha<tol)
        break;
    end
    
end
t_balance = toc(t_b);

display(sprintf(['\nNoise level: delta = %e\n' ...
    'Balancing principle returned alpha = %0.3e after %d iterations (%0.2f seconds).\n' ...
    'Final reconstruction error: %0.3e, reconstruction time: %0.2f seconds.'],...
    delta,alpha,iter,t_balance,err,t_ssn));

%% Compare with L2 fitting
opts.verb = 1;
display(sprintf('\nComputing optimal L2 reconstruction.'));

alrange = unique([linspace(alpha/10,alpha,51) linspace(alpha,alpha*10,51)]);
err_opt = 1e99;
for altest = alrange
    utest = l2nonlinfit2d(fem,Mf,y,altest,ue,opts);
    err2  = norm(utest-ue(:))/sqrt(nel);
    if err2 < err_opt
        err_opt = err2;
        alpha2  = altest;
        u2      = utest;
    end
end

%% show reconstruction
figure(4); fem.plot(u); title(['L1 reconstruction (\alpha=' num2str(alpha,'%0.2e') ')']);
shading interp; camlight headlight; lighting phong; axis tight; drawnow;
figure(5); fem.plot(u2); title(['L2 reconstruction (\alpha=' num2str(alpha2,'%0.2e') ')']);
shading interp; camlight headlight; lighting phong; axis tight; drawnow;
