function u = l2nonlinfit(fem,Mf,yd,alpha,ue,opts)
% U = L2NONLINFIT(FEM,MF,YD,ALPHA,UE,OPTS)
% solve nonlinear L2 data fitting problem using Newton method
%
% Input:  fem    -- finite element structure containing matrices
%                   (A(u),M,Y) and projection operators (P,P2)
%         Mf     -- right hand side with mass matrix already applied
%         yd     -- measured data
%         alpha  -- regularization parameter for L2 penalty
%         ue     -- exact solution (to evaluate reconstruction error)
%         opts   -- options structure containing (defaults)
%            maxit -- max. number of Newton iterations (20)
%            tol   -- tolerance for Newton iteration: norm of gradient (1e-6)
%            cgits -- max. number of BiCGstab iterations (1000)
%            cgtol -- tolerance for BiCGstab (1e-6)
%            verb  -- verbosity (0: no display)
% Output: u      -- reconstruction
%
% Christian Clason (christian.clason@uni-graz.at)
% Bangti Jin       (btjin@math.tamu.edu)
% February 27, 2011

%% Setup parameters
nel   = length(fem.x);     % problem size: number of elements

% Default parameters
if nargin <= 5 || isempty(opts)
    maxit = 20;             % max. number of semi-smooth Newton iterations
    tol   = 1e-6;           % tolerance for Newton iteration (norm of gradient)
    cgits = 1000;           % max. number of BiCGstab iterations
    cgtol = 1e-6;           % tolerance for BiCGstab
    verb  = 1;              % verbosity (0,1: no display, 2: only path-following, 3: also SSN)
else
    maxit = opts.maxit;
    tol   = opts.tol;
    cgits = opts.cgits;
    cgtol = opts.cgtol;
    verb  = opts.verb;
end

DD = fem.DD;     % stiffness matrix: regularization term
M  = fem.M;      % Mass matrix
Mx = fem.Mx;     % differentiated mass matrix
Y  = fem.Y;      % Y(v)*du assembles weighted mass matrix for piecewise constant du and linear v
P  = fem.P;      % projection to piecewise linear space

%% SSN loop

% initialize iterates
u = ones(nel,1);

iter = 0;

while (iter<maxit)
    
    A = fem.A(u);                  % differential operator
    y = A\(Mf);                    % state
    p = y-yd;                      % residual
    
    q = A\(M*p);                   % adjoint
    
    % precompute mass matrices for Hessian application
    yx = Mx*y; Myx = Y(yx);
    qx = Mx*q; Mqx = Y(qx);
    
    % gradient (preconditioning with DD^(-1))
    rhs  = -(alpha*u - DD\P(yx.*qx));
    nres = norm(rhs)/sqrt(nel);
    
    % Hessian
    C = @(x) applyHess(x,A,DD,M,Mx,P,yx,qx,Myx,Mqx,alpha);
    
    % Newton step
    [dx,flag] = bicgstab(C,rhs,cgtol,cgits); %#ok<NASGU>
    
    u = u + dx;
    
    
    if verb > 1
        fprintf('Newton iteration %d: |grad|=%e, |u-uex|=%e \n',iter,nres,norm(u-ue)/sqrt(nel));
    end
    
    if nres < tol
        break;
    end
    
    iter   = iter+1;
    
end

end % main function

function Hdu = applyHess(x,A,DD,M,Mx,P,yx,qx,Myx,Mqx,alpha)
dy = A\(Myx*x);                   % differential change in state
dq = A\(Mqx*x-(M*dy));            % differential change in adjoint

Hdu = alpha*x - DD\P(yx.*(Mx*dq)+qx.*(Mx*dy));
end % function applyHess
