function u = l2nonlinfit2d(fem,Mf,yd,alpha,ue,opts)
% U = L2NONLINFIT(FEM,MF,YD,ALPHA,UE,OPTS)
% solve nonlinear L2 data fitting problem using Newton method
%
% Input:  fem    -- finite element structure containing matrices
%                   (A(u),M,Y) and projection operators (P,P2)
%         Mf     -- right hand side with mass matrix already applied
%         yd     -- measured data
%         alpha  -- regularization parameter for L2 penalty
%         ue     -- exact solution (to evaluate reconstruction error)
%         opts   -- options structure containing (defaults)
%            maxit -- max. number of Newton iterations (20)
%            tol   -- tolerance for Newton iteration: norm of gradient (1e-6)
%            cgits -- max. number of BiCGstab iterations (1000)
%            cgtol -- tolerance for BiCGstab (1e-6)
%            verb  -- verbosity (0: no display)
% Output: u      -- reconstruction
%
% Christian Clason (christian.clason@uni-graz.at)
% Bangti Jin       (btjin@math.tamu.edu)
% February 27, 2011

%% Setup parameters
nel   = length(fem.xx(:));     % problem size: number of nodes

% Default parameters
if nargin <= 5 || isempty(opts)
    maxit = 20;             % max. number of semi-smooth Newton iterations
    tol   = 1e-6;           % tolerance for Newton iteration (norm of gradient)
    cgits = 1000;           % max. number of BiCGstab iterations
    cgtol = 1e-6;           % tolerance for BiCGstab
    verb  = 1;              % verbosity (0,1: no display, 2: only path-following, 3: also SSN)
else
    maxit = opts.maxit;
    tol   = opts.tol;
    cgits = opts.cgits;
    cgtol = opts.cgtol;
    verb  = opts.verb;
end

DD = fem.DDinv;  % stiffness matrix: regularization term (inverse for preconditioner)
M  = fem.M0;     % Mass matrix (Dirichlet)
Mx = fem.Mx;     % differentiated mass matrix
My = fem.My;     % differentiated mass matrix
MY = fem.Y;      % Y(v)*du assembles weighted mass matrix for piecewise constant du and linear v
P  = fem.P;      % projection from U_h to V_h

%% SSN loop

% initialize iterates
u = ones(nel,1);

iter = 0;

while (iter<maxit)
    
    A = fem.A(u);                 % differential operator
    try
        R = chol(A); Rt=R';       % precompute Cholesky factors; faster in 2D
        S = @(f) R\(Rt\f);        % (linearized) solution operator
    catch notspd %#ok<NASGU>
        S = @(f) A\f;             % fallback if numerically semidefinite
    end
    
    y = S(Mf);                    % state
    p = y-yd;                     % residual
    
    q = S(M*p);                   % adjoint
    
    % precompute mass matrices for Hessian application
    yx = Mx*y; yy = My*y; Mgy = MY(yx,yy);
    qx = Mx*q; qy = My*q; Mgq = MY(qx,qy);
    
    % gradient: preconditioning with DD^{-1}
    rhs  = -(alpha*u - DD(P*(yx.*qx+yy.*qy)));
    nres = norm(rhs)/sqrt(nel);
    if nres < tol
        break
    end
    
    % Hessian
    C = @(x) applyHess(x,S,DD,M,Mx,My,P,yx,yy,qx,qy,Mgy,Mgq,alpha);
    
    % Newton step
    [dx,flag] = bicgstab(C,rhs,cgtol,cgits); %#ok<NASGU>
    
    u = u + dx;
    
    if verb > 1
        fprintf('Newton iteration %d: |grad|=%e, |u-uex|=%e \n',iter,nres,norm(u-ue(:))/sqrt(nel));
    end
    
    if nres < tol
        break;
    end
    
    iter   = iter+1;
    
end

end % main function


%% Application of Hessian
function Hdu = applyHess(x,S,DD,M,Mx,My,P,yx,yy,qx,qy,Mgy,Mgq,alpha)
dy = S(Mgy*(x));                  % differential change in state
dq = S(Mgq*(x)-(M*dy));           % differential change in adjoint

Hdu = alpha*x - DD(P*(yx.*(Mx*dq)+qx.*(Mx*dy) + yy.*(My*dq)+qy.*(My*dy)));
end % function applyHess
