function fem = setupFEM(nel)
% FEM = SETUPFEM(NEL)
% setup 1D FEM structure
% Input:  nel -- number of elements
% Output: fem -- structure containing
%            x  -- nodes
%            xc -- cell centers (used for plotting)
%            dx -- mesh size
%            A  -- function A(u) returning stiffness matrix (-\Delta + u)
%         M,M0  -- mass matrix for right hand side in V_h (apply Dirichlet conditions)
%            Y  -- function Y(y) returning weighted mass matrix
%            P  -- projection V_h to U_h: trapezoidal rule
%           DD  -- regularization term
%
% Christian Clason (christian.clason@uni-graz.at)
% Bangti Jin       (btjin@math.tamu.edu)
% February 27, 2011

%% set up grid
n  = nel+1;              % number of grid points
x  = linspace(0,1,n)';   % spatial grid points
dx = x(2)-x(1);          % spatial grid step size
xc = x(2:end)-dx/2;      % cell centers

%% set up matrices
e = ones(n,1);
% Laplace with homogeneous Neumann conditions
K = spdiags([-e 2*e -e]/dx,-1:1,n,n); K(1,1:2) = [1 -1]/dx; K(end,end-1:end) = [-1 1]/dx;

% mass matrix for piecewise linear right hand side
M = spdiags([e 4*e e]*dx/6,-1:1,n,n); M(1,1) = dx/3; M(end,end) = dx/3;

%% FEM structure: grid, stiffness & mass matrices, projections etc.
fem.x  = x;                              % grid
fem.xc = xc;                             % cell centers
fem.dx = dx;                             % mesh size

fem.A  = @(u) assembleA(u,dx);           % differential operator (div(u grad(y))
fem.DD = K+M;                            % regularization term: -Delta + I
fem.M  = M;
fem.M0 = M; fem.M0(1,1:2)=0; fem.M0(end,end-1:end)=0;  % mass matrix: apply Dirichlet cond.
fem.Mx = spdiags([-e e]/dx,0:1,nel,n);   % differentiation matrix
fem.Y  = @(y) assembleY(y);              % weighted mass matrix: Y(y)*du = <y' du,v'> for du in U_h
fem.P  = @(y) dx/2*([y;0]+[0;y]);        % apply stiffness matrix: trapezoidal rule

%% regularization term: -\Delta + I, homogeneous Neumann
fem.DD = K+M;                            % regularization term: -Delta + I

end % function assembleFEM

%% Assemble stiffness matrix <u nabla y, nabla v> for u in V_h
function K = assembleA(u,dx)
nel = length(u)-1; % number of elements

Ke = [1 -1 -1 1]'/dx;  % elemental mass matrix <nabla phi_i,nabla phi_j>

ent = 4*nel;
row = zeros(ent,1);
col = zeros(ent,1);
val = zeros(ent,1);
ind = 1;
for el = 1:nel % Dirichlet conditions
    gl      = ind:(ind+3);
    row(gl) = el + [0 1 0 1];
    col(gl) = el + [0 0 1 1];
    val(gl) = (u(el)+u(el+1))/2*Ke;
    ind     = ind+4;
end
K = sparse(row,col,val);

% Dirichlet conditions
K(1,:) = 0; K(1,1) = 1; K(end,:) = 0; K(end,end) = 1;
end %function assembleU

%% Assemble weighted mass matrix: Y(y)*du = <grad y du,grad v> for du in U_h
function MY = assembleY(My)
n = length(My);
Myp = [0;My(1:end)]; Mym = [My(1:end);0];
MY = spdiags([Mym Myp-Mym -Myp]/2,-1:1,n+1,n+1);
MY(1,1:2) = 0; MY(end,end-1:end) = 0; % enforce Dirichlet conditions
end % function assembleY
