function fem = setupFEM2d(n)
% FEM = SETUPFEM2D(NEL)
% setup 2D FEM structure (uniform triangular grid)
% Input:  n   -- number of grid points in 1D
% Output: fem -- structure containing
%          nel   -- number of elements
%          xx,yy -- nodes
%          h2    -- mesh constant: 2*area(T)
%          A     -- function A(u) returning stiffness matrix (-\Delta + u)
%          M,M0  -- mass matrix for right hand side in V_h (with Dirichlet conditions)
%          Y     -- function Y(y) returning weighted mass matrix
%          P     -- projection to piecewise linear space: trapezoidal rule
%          plot  -- plot cell-wise constant functions
%
% Christian Clason (christian.clason@uni-graz.at)
% Bangti Jin       (btjin@math.tamu.edu)
% February 27, 2011

%% Set up grid: uniform triangular mesh
x       = linspace(0,1,n)';      % spatial grid points (uniform in x,y)
h       = (x(2)-x(1));           % grid size
[xx,yy] = meshgrid(x);           % coordinates of nodes
nel     = 2*(n-1)^2;             % number of elements

tri     = zeros(nel,3);          % triangulation
ind = 1;
for i = 1:n-1
    for j = 1:n-1
        node         = (i-1)*n+j+1;              % two triangles meeting at node
        tri(ind,:)   = [node node-1 node+n];     % first triangle (lower left)
        tri(ind+1,:) = [node+n-1 node+n node-1]; % second triangle (upper right)
        ind = ind+2;
    end
end
bdnod = find((xx == 0) | (xx == 1) | (yy == 0) | (yy == 1)); % boundary nodes

fem.plot = @(f) surf(xx,yy,reshape(f,n,n));

%% stiffness and mass matrices
[K,M]     = assembleFEM(tri,h^2);
[Mx,My,P] = assembleMx(tri,h);

%% FEM structure: grid, stiffness & mass matrices, projections etc.
fem.nel = nel;                         % number of elements
fem.xx  = xx;                          % grid
fem.yy  = yy;                          % grid
fem.h2  = h^2;                         % mesh constant: 2*area(T)
fem.A   = @(u) assembleA(u,tri,bdnod); % differential operator
fem.M   = M;
fem.M0  = M; fem.M0(bdnod,:) = 0;      % mass matrix: apply Dirichlet cond.
fem.Mx  = Mx; fem.My = My;             % differentiated mass matrix
fem.Y   = @(fx,fy) assembleY(fx,fy,tri,bdnod,h); % weighted mass matrix: Y(y)*du = <y du,v>
fem.P   = P;                           % projection from piecewise constants to linears

%% regularization term: -\Delta + I, homogeneous Neumann
fem.DD    = K+M;                       % regularization term
R = chol(fem.DD); Rt = R';
fem.DDinv = @(x) R\(Rt\x);             % apply DD^(-1): preconditioner

end % function setupFEM2D


%% Assemble stiffness, mass matrix
function [K,M] = assembleFEM(t,h2)
nel = size(t,1);                   % number of elements

Ke = 1/2*[2 -1 -1 -1 1 0 -1 0 1]'; % elemental stiffness matrix <phi_i',phi_j'>
Me = h2/24 * [2 1 1 1 2 1 1 1 2]'; % elemental mass matrix <phi_i,phi_j>

ent = 9*nel;
row = zeros(ent,1);
col = zeros(ent,1);
valk  = zeros(ent,1);
valm  = zeros(ent,1);

ind=1;
for el=1:nel
    ll      = ind:(ind+8);         % local node indices
    gl      = t(el,:);             % global node indices
    cg      = gl([1;1;1],:); rg = gl';
    rg      = rg(:,[1 1 1]);
    row(ll) = rg(:);
    col(ll) = cg(:);
    valk(ll)= Ke;
    valm(ll)= Me;
    ind     = ind+9;
end
K  = sparse(row,col,valk);
M  = sparse(row,col,valm);
end % function assembleFEM

%% Assemble differentiation, projection matrix
function [Mx,My,P] = assembleMx(t,h)
nel = size(t,1);                   % number of elements

mx = 1/h*[ -1 0 1 1 0 -1]';
my = 1/h*[ 1 -1 0 -1  1 0]';

ent   = 3*nel;
row   = zeros(ent,1);
col   = zeros(ent,1);
valmx = zeros(ent,1);
valmy = zeros(ent,1);

ind=1;
for el=1:2:nel
    ll       = ind:(ind+5);        % global node indices: left triangle
    row(ll)  = [el el el el+1 el+1 el+1];
    col(ll)  = t(el:el+1,:)';
    valmx(ll)= mx;
    valmy(ll)= my;
    ind      = ind+6;
end
Mx = sparse(row,col,valmx);
My = sparse(row,col,valmy);
P  = sparse(col,row,h^2/6);
end % function assembleMx

%% Assemble parameter dependent stiffness matrix <u grad y,grad v> for u in V_h
function A = assembleA(u,t,bdnod)
nel = size(t,1);                   % number of elements

Ke = 1/6*[2 -1 -1 -1 1 0 -1 0 1]'; % elemental stiffness matrix <phi_i',phi_j'>/3

ent = 9*nel;
row = zeros(ent,1);
col = zeros(ent,1);
val = zeros(ent,1);

ind=1;
for el=1:nel
    ll      = ind:(ind+8);         % local node indices
    gl      = t(el,:);             % global node indices
    cg      = gl([1;1;1],:); rg = gl';
    rg      = rg(:,[1 1 1]);
    row(ll) = rg(:);
    col(ll) = cg(:);
    val(ll) = sum(u(gl))*Ke;
    ind     = ind+9;
end
A = sparse(row,col,val);

A(bdnod,:) = 0;  A(:,bdnod) = 0;   % Dirichlet conditions
for j=1:length(bdnod)
    A(bdnod(j),bdnod(j)) = 1;  %#ok<SPRIX>
end

end % function assembleA

%% Assemble weighted mass matrix: Y(y)*du = <grad y du,grad v> for du in V_h
function Y = assembleY(fx,fy,t,bdnod,h)
nel = size(t,1);                    % number of elements

mx = h/6*[-1 0 1  -1 0 1  -1 0 1]'; % <phi_i, d_x phi_j>
my = h/6*[1 -1 0  1 -1 0  1 -1 0]'; % <phi_i, d_y phi_j>

ent = 9*nel;
row = zeros(ent,1);
col = zeros(ent,1);
val = zeros(ent,1);

ind=1;
for el=1:nel
    ll      = ind:(ind+8);          % local node indices
    gl      = t(el,:);              % global node indices
    cg      = gl([1;1;1],:); rg = gl';
    rg      = rg(:,[1 1 1]);
    row(ll) = rg(:);
    col(ll) = cg(:);
    val(ll) = (-1)^(el+1)*(mx*fx(el)+my*fy(el)); % <y_x phi_i, phi_j'>_K
    ind     = ind+9;
end

Y = sparse(row,col,val);

Y(bdnod,:) = 0;                    % Dirichlet conditions

end % function assembleY
