function u = l1nonlinfit_ssn2d(fem,Mf,yd,alpha,ue,opts)
% U = L1NONLINFIT_SSN2D(FEM,MF,YD,ALPHA,UE,OPTS)
% solve nonlinear L1 data fitting problem using path-following semi-smooth
%  Newton method (using Cholesky pre-factorization for 2D problems)
%
% Input:  fem    -- finite element structure containing matrices
%                   (A(u),M,Y), projection operators (P), boundary nodes (obs)               (obs)
%         Mf     -- right hand side with mass matrix already applied
%         yd     -- measured data
%         alpha  -- regularization parameter for L2 penalty
%         ue     -- exact solution (to evaluate reconstruction error)
%         opts   -- options structure containing (defaults)
%            beta0 -- initial penalty parameter for p (1)
%            bred  -- reduction factor for beta (1/2)
%            bmin  -- minimal beta value (1e-12)
%            c     -- penalty parameter for box constraint (1e3)
%            maxit -- max. number of semi-smooth Newton iterations (20)
%            tol   -- tolerance for Newton iteration: norm of gradient (1e-6)
%            cgits -- max. number of BiCGstab iterations (1000)
%            cgtol -- tolerance for BiCGstab (1e-6)
%            verb  -- verbosity (0,1: no display, 2: only path-following, 3: also SSN)
% Output: u      -- reconstruction
%
% Christian Clason (christian.clason@uni-graz.at)
% Bangti Jin       (btjin@math.tamu.edu)
% February 27, 2011

%% Setup parameters
nel   = size(fem.xc,1);     % problem size: number of elements

% Default parameters
if nargin <= 5 || isempty(opts)
    beta  = 1;              % initial penalty parameter for p
    bred  = 1/2;            % reduction factor for beta
    bmin  = 1e-12;          % minimal beta value
    maxit = 20;             % max. number of semi-smooth Newton iterations
    tol   = 1e-6;           % tolerance for Newton iteration (norm of gradient)
    cgits = 1000;           % max. number of BiCGstab iterations
    cgtol = 1e-6;           % tolerance for BiCGstab
    verb  = 1;              % verbosity (0,1: no display, 2: only path-following, 3: also SSN)
else
    beta  = opts.beta;
    bred  = opts.bred;
    bmin  = opts.bmin;
    maxit = opts.maxit;
    tol   = opts.tol;
    cgits = opts.cgits;
    cgtol = opts.cgtol;
    verb  = opts.verb;
end

P   = fem.P;      % Projection onto piecewise constants for linear functions
M   = fem.M;      % Mass matrix for piecewise constant on boundary
Y   = fem.Y;      % Y(v)*du assembles weighted mass matrix for piecewise constant du and linear v
obs = fem.obs;    % restriction to nodes on observation boundary

%% SSN loop

% initialize iterates
u = ones(nel,1);

while beta>bmin
    
    iter   = 0;
    update = 1;
    nres   = 1;
    uold   = u;
    
    while (iter<maxit)
        
        A = fem.A(u);             % differential operator
        try
            R = chol(A); Rt=R';   % precompute Cholesky factors; faster in 2D
            S = @(f) R\(Rt\f);    % (linearized) solution operator
        catch notspd %#ok<NASGU>
            S = @(f) A\f;         % fallback if numerically semidefinite
        end
        
        y = S(Mf);                % state
        res = y(obs)-yd;          % residual
        
        % active sets
        Ap = (res - beta)>0;      
        Am = (res + beta)<0;      
        As = Ap+Am;
        
        p = 0*y;
        p(obs) = (1-As).*res/beta + (Ap - Am);
        
        q  = S(-(M*p));           % adjoint
        
        My = Y(y);
        Mq = Y(q);                % precompute mass matrices for Hessian application
        
        % gradient
        rhs  = -(alpha*u + P(y.*q));
        nres = norm(rhs)/sqrt(nel);
        
        % Hessian
        C = @(x) applyHess(x,S,P,M,obs,y,q,My,Mq,As,alpha,beta);
        
        % Newton step
        [dx,flag] = bicgstab(C,rhs,cgtol,cgits); %#ok<NASGU>
        
        u = u + dx;
        
        if iter == 0
            update = 1;
        else
            update = nnz((Ap-Ap_old)|(Am-Am_old));
            if verb > 2
                fprintf('SSN iteration %d: %d points changed in active set, beta=%1.2e, |grad|=%e\n', iter, update,beta,nres);
            end
        end
        iter   = iter+1;
        Ap_old = Ap;
        Am_old = Am;
        
        if ~update && nres<tol    % terminate if active sets coincide and norm of gradient small
            break
        end
        
    end
    
    if verb > 1
        fprintf('beta=%1.2e: %d points changed in active set after %d iterations, |grad| = %e, |u-uex|=%e \n',beta,update,iter,nres,norm(u-ue)/sqrt(nel));
    end
    
    if update || nres>tol         % iterate not feasible: terminate and return previous iterate
        u = uold;
        break
    else                          % reduce beta, continue
        beta = beta*bred;
    end
    
end

end % main function


%% Application of Hessian
function Hdu = applyHess(x,S,P,M,obs,y,q,My,Mq,As,alpha,beta)
dy = S(-My*x);                    % differential change in state
dp = dy*0;
dp(obs) = (1-As).*dy(obs)/beta;
dq = S(-Mq*x-M*dp);               % differential change in adjoint

Hdu = alpha*x + P(y.*dq+q.*dy);
end % function applyHess
