% L1NONLINFIT_TEST2D
% Test script for nonlinear L^1 data fitting using automatic parameter
% choice and path-following semismooth Newton method (2D). Solves
%  min_u |S(u)-yd|_L1 + alpha/2 (|u|_L2)^2, where S(u)=:y satisfies
%   <\nabla y, \nabla v > + <u y,v>_gamma_i = <f,v> for all v in H^1
%
% Christian Clason (christian.clason@uni-graz.at)
% Bangti Jin       (btjin@math.tamu.edu)
% February 27, 2011

clear, close all
%% parameters
n     = 64;        % problem size
d_per = 0.3;       % percentage of corrupted data points
verb  = 2;         % verbosity (0: no display, 1: balancing only, 2: also path-following, 3: also SSN)

% ... for balancing principle
alpha0 = 1;        % initial guess for parameter alpha
sigma  = 1.03;     % balancing factor
tol    = 1e-4;     % terminate if relative change in alpha less than tol

% ... for path-following semi-smooth Newton method
opts.beta  = 1;              % initial penalty parameter for p
opts.bred  = 0.5;            % reduction factor for beta
opts.bmin  = 1e-12;          % minimal beta value
opts.maxit = 20;             % max. number of semi-smooth Newton iterations
opts.tol   = 1e-6;           % tolerance for Newton iteration (norm of gradient)
opts.cgits = 100;            % max. number of BiCGstab iterations
opts.cgtol = 1e-6;           % tolerance for BiCGstab
opts.verb  = verb;           % verbosity

%% Data
fem = setupFEM2d(n);         % FEM setup
xx  = fem.xx; yy = fem.yy;   % grid
xc  = fem.xc;                % cell centers on Robin boundary (for plotting)
Mf  = fem.M*(-4+xx(:));      % right hand side (pre-multiply mass matrix)
ue =  1+xc;                  % true solution
ye  = fem.A(ue)\(Mf);        % exact data

% noisy data (additive random-valued impulsive noise)
nel    = length(xc);         % number of element (unknown boundary)
y      = ye;
ind    = find(rand(n^2,1) < d_per);
y(ind) = ye(ind) + max(abs(ye)) * randn(length(ind),1);
ye     = ye(fem.obs); y = y(fem.obs);

nnod   = length(ye);         % number of nodes (obs. boundary)
delta  = norm(y-ye,1)/nnod;   % compute noise level

% show the noisy data
figure(1), plot(1:length(ye),ye,'k',1:length(ye),y,'r');
title('Data'); legend('exact',['noisy (\delta=' num2str(delta,'%10.2e') ')']);

%% Balancing principle loop
t_b = tic;

alpha = alpha0;
for iter = 1:10
    alpha1 = alpha;
    
    % compute u_\alpha
    t_s    = tic;
    u      = l1nonlinfit_ssn2d(fem,Mf,y,alpha,ue,opts);
    t_ssn  = toc(t_s);
    
    % update alpha
    yalpha = fem.A(u)\Mf;
    alpha = (sigma-1)*(norm(yalpha(fem.obs)-y(:),1)/nnod)/(norm(u(:))^2/(2*nel));
    err   = norm(u-ue)/sqrt(nel);
    
    if opts.verb > 0
        figure(99); plot(xc,u,xc,ue);
        display(sprintf('Iteration %d: t = %g seconds, alpha = %g, err= %g',...
            iter,t_ssn,alpha,err));
    end
    
    % terminate if relative change in alpha too small
    if (abs(alpha-alpha1)/alpha<tol)
        break;
    end
    
end
t_balance = toc(t_b);

display(sprintf(['\nNoise level: delta = %e\n' ...
    'Balancing principle returned alpha = %0.3e after %d iterations (%0.2f seconds).\n' ...
    'Final reconstruction error: %0.3e, reconstruction time: %0.2f seconds.'],...
    delta,alpha,iter,t_balance,err,t_ssn));

%% Compare with L2 fitting
opts.verb = 1;
display(sprintf('\nComputing optimal L2 reconstruction.'));

alrange = unique([linspace(alpha*10,alpha*100,51) linspace(alpha*100,alpha*1000,51)]);
err_opt = 1e99;
for altest = alrange
    utest = l2nonlinfit2d(fem,Mf,y,altest,ue,opts);
    err2  = norm(utest-ue)/sqrt(nel);
    if err2 < err_opt
        err_opt = err2;
        alpha2  = altest;
        u2      = utest;
    end
end

%% show reconstruction
figure(2),plot(xc,u,'r',xc,ue,'k',xc,u2,'b--');drawnow
title('Reconstruction');
legend(['L^1 (\alpha=' num2str(alpha,'%0.2e') ')'],'exact',...
    ['L^2 (\alpha=' num2str(alpha2,'%0.2e') ')'],'Location','SouthEast');
